/* Name: main.c
 * Project: HID-Test
 * Author: Christian Starkjohann
 * Creation Date: 2006-02-02
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 299 2007-03-29 17:07:19Z cs $
 */

//#define F_CPU   12000000L    /* evaluation board runs on 4MHz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"

uchar enc_states[] = {0,2,1,0,1,0,0,2,2,0,0,1,0,1,2,0};
uchar encoder1status;
uchar encoder2status;

/* ----------------------- hardware I/O abstraction ------------------------ */

/* pin assignments:
PB0	Key 2
PB1	Key 7
PB2	Key 3
PB3	Bootloader jumper & Key 15
PB4	Key 1
PB5	Status LED

PC0	Enc2 Left
PC1	Key 12
PC2	Key 10
PC3	Enc1 Right
PC4	Enc2 Right
PC5	Enc1 Left

PD0	Key 11
PD1	Key 6
PD2	USB+ (int0)
PD3	Key 5
PD4	Key 9
PD5	Key 4
PD6	Key 8
PD7	USB-
*/

static void hardwareInit(void)
{
uchar	i, j;

    DDRB = 0x20;       /* all pins input except led */

    PORTD = 0x00;   /* no pull-ups */
    DDRD = 0x14;    /* 0001 0101 bin: all pins input except USB (-> USB reset) */
	j = 0;
	while(--j){     /* USB Reset by device only required on Watchdog Reset */
		i = 0;
		while(--i); /* delay >10ms for USB reset */
	}
    DDRD = 0x00;    /* 0000 0010 bin: remove USB reset condition */
    /* configure timer 0 for a rate of 12M/(1024 * 256) = 45.78 Hz (~22ms) */

    PORTB |= 0x1F; // Activate all pull-ups.
    PORTC |= 0x3F;
    PORTD |= 0x7F;
    TCCR0B = 5;      /* timer 0 prescaler: 1024 */
    encoder1status = 0;
    encoder2status = 0;
}

/* ------------------------------------------------------------------------- */

#define NUM_KEYS    17

/* The following function returns an index for the first key pressed. It
 * returns 0 if no key is pressed.
 */
static uchar    keyPressed(void)
{
uchar   x;

    x = PINB;
    if ((x & 0x1F) != 0x1F) { // Key pressed on port B;
      PORTB |= 0x20; // Turn on LED
      if (!(x & 1)) return 2;
      if (!(x & 2)) return 7;
      if (!(x & 4)) return 3;
      if (!(x & 8)) return 15;
      if (!(x & 0x10)) return 1;
    }

    x = PINC;
    encoder1status <<= 2;                   //remember previous state
    encoder1status |= ((x & 0x08) >> 2) | ((x & 0x20) >> 4);  //add current state
    if ( enc_states[( encoder1status & 0x0f )]) {
      PORTB |= 0x20; // Turn on LED
      if ( enc_states[( encoder1status & 0x0f )] == 1) return 13;
      if ( enc_states[( encoder1status & 0x0f )] == 2) return 14;
    }
    encoder2status <<= 2;                   //remember previous state
    encoder2status |= ((x & 0x10) >> 3) | (x & 1);  //add current state
    if ( enc_states[( encoder2status & 0x0f )]) {
      PORTB |= 0x20; // Turn on LED
      if ( enc_states[( encoder2status & 0x0f )] == 1) return 16;
      if ( enc_states[( encoder2status & 0x0f )] == 2) return 17;
    }
    if ((x & 0x2E) != 0x2E) { // Key pressed on port C;
      PORTB |= 0x20; // Turn on LED
      if (!(x & 2)) return 12;
      if (!(x & 4)) return 10;
      if (!(x & 8)) return 13;
      if (!(x & 0x20)) return 14;
    }
    x = PIND;
    if ((x & 0x7B) != 0x7B) { // Key pressed on port D;
      PORTB |= 0x20; // Turn on LED
      if (!(x & 1)) return 11;
      if (!(x & 2)) return 6;
      if (!(x & 8)) return 5;
      if (!(x & 0x10)) return 9;
      if (!(x & 0x20)) return 4;
      if (!(x & 0x40)) return 8;
    }
    PORTB &= ~0x20; // Turn off LED
    return 0;
}
/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

PROGMEM char usbHidReportDescriptor[35] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

#define KEY_A       4
#define KEY_B       5
#define KEY_C       6
#define KEY_D       7
#define KEY_E       8
#define KEY_F       9
#define KEY_G       10
#define KEY_H       11
#define KEY_I       12
#define KEY_J       13
#define KEY_K       14
#define KEY_L       15
#define KEY_M       16
#define KEY_N       17
#define KEY_O       18
#define KEY_P       19
#define KEY_Q       20
#define KEY_R       21
#define KEY_S       22
#define KEY_T       23
#define KEY_U       24
#define KEY_V       25
#define KEY_W       26
#define KEY_X       27
#define KEY_Y       28
#define KEY_Z       29
#define KEY_1       30
#define KEY_2       31
#define KEY_3       32
#define KEY_4       33
#define KEY_5       34
#define KEY_6       35
#define KEY_7       36
#define KEY_8       37
#define KEY_9       38
#define KEY_0       39

#define KEY_PLUS    0x57
#define KEY_MINUS   0x56

#define KEY_MINUS_UNDERSCORE    0x2D
#define KEY_EQUALS_PLUS		0x2E

#define RIGHT_ARR   0x4F
#define LEFT_ARR    0x50
#define DOWN_ARR    0x51
#define UP_ARR      0x52

#define ENTER       0x28
#define SPACE       0x2C
#define BACKSPACE   0x2A
#define TAB         0x2B
#define HOME        0x4A
#define ESC         0x29
#define KEY_APP     0x65 // Windows key

#define KEY_F1      0x3A
#define KEY_F2      0x3B
#define KEY_F3      0x3C
#define KEY_F4      0x3D
#define KEY_F5      0x3E
#define KEY_F6      0x3F
#define KEY_F7      0x40
#define KEY_F8      0x41
#define KEY_F9      0x42
#define KEY_F10     0x43
#define KEY_F11     0x44
#define KEY_F12     0x45

#define VOLUMEUP    0x80
#define VOLUMEDOWN  0x81

static const uchar  keyReport[NUM_KEYS + 1][2] PROGMEM = {
/* none */  {0, 0},                     /* no key pressed */
/*  1 */    {0, LEFT_ARR},
/*  2 */    {0, RIGHT_ARR},
/*  3 */    {0, UP_ARR},
/*  4 */    {0, DOWN_ARR},
/*  5 */    {MOD_ALT_LEFT, TAB},
/*  6 */    {0, SPACE},
/*  7 */    {0, KEY_B},
/*  8 */    {0, KEY_F8},
/*  9 */    {0, KEY_F7},
/* 10 */    {0, KEY_APP},
/* 11 */    {0, KEY_H},
/* 12 */    {MOD_CONTROL_LEFT | MOD_SHIFT_LEFT, ESC},
/* Enc1L */ {0, VOLUMEDOWN},
/* Enc1R */ {0, VOLUMEUP},
/* 15 */    {0, KEY_Y},
/* Enc2L */ {0, LEFT_ARR},
/* Enc2R */ {0, RIGHT_ARR},
};

static void buildReport(uchar key)
{
/* This (not so elegant) cast saves us 10 bytes of program memory */
    *(int *)reportBuffer = pgm_read_word(keyReport[key]);
}

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport(keyPressed());
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

/* ------------------------------------------------------------------------- */

int	main(void)
{
uchar   key, lastKey = 0, keyDidChange = 0;
uchar   idleCounter = 0;

	wdt_enable(WDTO_2S);
    hardwareInit();
	odDebugInit();
	usbInit();
	sei();
    DBG1(0x00, 0, 0);
	for(;;){	/* main event loop */
		wdt_reset();
		usbPoll();
        key = keyPressed();
        if(lastKey != key){
            lastKey = key;
            keyDidChange = 1;
        }
        if(TIFR0 & (1<<TOV0)){   /* 22 ms timer */
            TIFR0 = 1<<TOV0;
            if(idleRate != 0){
                if(idleCounter > 4){
                    idleCounter -= 5;   /* 22 ms in units of 4 ms */
                }else{
                    idleCounter = idleRate;
                    keyDidChange = 1;
                }
            }
        }
        if(keyDidChange && usbInterruptIsReady()){
            keyDidChange = 0;
            /* use last key and not current key status in order to avoid lost
               changes in key status. */
            buildReport(lastKey);
            usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        }
	}
	return 0;
}

/* ------------------------------------------------------------------------- */
